#!/usr/bin/env python
#
# A part of this script comes from the Cambridge Crystallographic Data Centre
# This script can be used for any purpose without limitation subject to the
# conditions at http://www.ccdc.cam.ac.uk/Community/Pages/Licences/v2.aspx
#
# This permission notice and the following statement of attribution must be
# included in all copies or substantial portions of this script.
#
#ChiPi can be used for any purpose and is working with python 2 and the v2.3 of the CCDC Python API 
#For any information, please contact: simon.clevers@gmail.com
#
#Disclaimer: part of the code are just attempt for future implementations, therefore all functionnalities are not used
#
#
#
#
'''
Provide information on a set of structures in the CSD.

This script takes as input a gcd file (a text file with CSD refcodes) and
writes out the in csv format information about the chirality of the crystal (achiral, chiral, 
racemic, scalemic, meso,diast). It also returns the molecular point group (only for chiral molecule)
and performs overlay comparison between each molecules in structure having Z'>1
 

'''
from enumChipi import EnumInfo as ei
from libChipi import StructureMatching as sm
import os
import sys
import time
import itertools
from itertools import chain, combinations
import numpy as np
from numpy import *
from ccdc.io import EntryReader, CrystalReader
from ccdc import io
from ccdc.descriptors import MolecularDescriptors as md, GeometricDescriptors as gd
import io
import multiprocessing as mp
import psutil

sys.stdout = io.TextIOWrapper(sys.stdout.detach(), encoding = 'utf-8')
sys.stderr = io.TextIOWrapper(sys.stderr.detach(), encoding = 'utf-8')
 
# Print
ARRAY_LEN = 4
NUMBER_MAX_ENTITY = 16
NUMBER_AFTER_COMMA = 4
NUMBER_OF_USEFULL_ENTITY = 4

# Algo
STOP = "stop"
FOR_RMSD = "RMSD"
OTHER = "OTHER"

NORMALISE = 0
NO_NORMALISE = 1

RMSD_INDEX = 1
TANIMOTO_INDEX = 2

# Main
INPUT_FILE = 1
RETURN_FILE = 2

# returnFile
TEMP = "return/temp"
TYPE = ".txt"
SEPARATORRMSD = '|'
SEPARATOR2RMSD = "|-|"
RETURN_FILE_LOCATION = "return/result"

# Return values
FUN_SUCCESS = 0
FUN_FAILURE = 1

class Writer(object):

  def __init__(self, infile):
    self.queue = mp.Queue()
    try:
      self.rdr = EntryReader(infile, format='identifiers')
    except RuntimeError:
      print('Failed to read input file %s!' % infile)
      exit(FUN_FAILURE)

  def printhelp(self):
    print("\nusage : python chipiVU.py input_file [output_file]")
    print("\ninput_file :\n\tLocation of a gcd file of required refcodes")
    print("\n[output_file] :\n\tLocation of the return file")

  def testType(self, v):
    return type(v) == float 

  def printArray2(self, array):
    s = ""
    for i in range(0, len(array)):
      s = s + SEPARATOR2RMSD if i != 0 else ''
      s = s + self.printArray(array[i], FOR_RMSD)
    return s

  def printArray(self, array, types):
    s = ""
    start = 0
    for j in range(start, len(array)):
      
      s = s + ((SEPARATORRMSD if types == FOR_RMSD else ',') if j != start else '')
      s = s + ("\"" if types == OTHER and j != start else "") + str(round(array[j], NUMBER_AFTER_COMMA) if (self.testType(array[j])) else array[j])
      s = s + ("\"" if types == OTHER and j != len(array) - 1 else "")

    return s

  def analyse(self):

    fw = open(RETURN_FILE_LOCATION + TYPE, 'w')
    fw.write(self.csv_header())
    t = time.time() 
    # if __name__ == '__main__':
      # procs = list()
      # n_cpus = psutil.cpu_count()
      # for e in self.rdr:
        # p = mp.Process(target=self.csv_line, args=(e, self.queue))
        # p.start()
        # procs.append(p)
      # while len(procs) > 0 :
        # if self.queue.empty() is False:
          # nq = self.queue.get()
          # fw.write('\n')
          # fw.write(nq)
          # procs[0].join()
          # procs.remove(procs[0])
      # for p in procs:
        # p.join()

    # print("monoproccessing")

    for e in self.rdr:
      fw.write('\n')
      fw.write(self.csv_line(e, None))
    fw.close()
    print("Calculation completed - time to write all : ",time.time() - t, " seconds")

  def csv_header(self):
    s = (','.join(ei.getAllValueInfo(ei.Info)))
    return s

  def assignAndAddH(self, comp, types):
    comp.assign_bond_types('Unknown')#assign unknown bond
    comp.add_hydrogens()#add hydrogen atom 
    if (types == NORMALISE):
      comp.normalise_atom_order()#normalise the atom order
      comp.normalise_labels()#verifies if each label is unique

  def c_StructureMatching(self, listOfList, mol_for_overlay, compo):
    structMatch = sm.StructureMatching()
    for i in listOfList:
      for mol_ab in i[0]:
        mol_1 = mol_ab[0] if (type(i[0]) == type(itertools.combinations(mol_for_overlay,2))) else compo[mol_ab[0]]
        self.assignAndAddH(mol_1, NO_NORMALISE)
        mol_2 = mol_ab[1] if (type(i[0]) == type(itertools.combinations(mol_for_overlay,2))) else compo[mol_ab[1]]
        self.assignAndAddH(mol_2, NO_NORMALISE)
        try:
          listOfBool = [False, True]
          listIndiceRmsdTanimoto = [RMSD_INDEX, TANIMOTO_INDEX]
          listr = []
          for j in listOfBool:
            r = md.overlay_rmsds_and_transformation(mol_1, mol_2, atoms=None,invert=True, rotate_torsions=i, with_symmetry=True)
            for k in listIndiceRmsdTanimoto:
              i[k].append(round(r[k], NUMBER_AFTER_COMMA))
            listr.append(r[3])

          rstructMatch = []
          for j in listr:
            rsm = structMatch.matchStruct(j)
            for k in rsm:
              rstructMatch.append(k)
        except RuntimeError as exp:
          warningc='Problem in overlay calculation'

  def c_pointGroup(self, compo, v, point_group_of_chiral_mol):
    try:
        compo[v[0]].remove_hydrogens()
        N_point_group=len(compo[v[0]].atoms)
    except IndexError as exp:
        warningc='Problem in point group determination'
        N_point_group=1000
    
    if 3<N_point_group<125:#the molecule is limited to 125 atoms (H excludes)- you can change this number but the calculation time will strongly increase.
      tempo=md.point_group_analysis(compo[v[0]])
      point_group_of_chiral_mol.append(str(tempo[1]))

    else:
      if N_point_group!=1000:
        warningc="Partially treated"
        

  def csv_line(self, e, queue):
    print(os.getpid(), "start")
    start_time = time.time()
    Fail=False
    formal_charge=False
    H_count=0
    C_count=0
    warning="error"
    warningb="NA"
    warningc="NA"
    krypto="NO"
    diast="NO"
    AU='ERROR_not_treated'
    final="error"
    crystal=e.crystal
    mol=e.molecule        
    chiral_point_group=['C1', 'C2', 'C3','D2','D3',[],'pg']
    cmax=len(mol.components)#calculate the number of residue in the AU
    Rcomp=np.array([],np.int)#to store the number of R atom in each residue
    Scomp=np.array([],np.int)#to store the number of S atom in each residue
    Mcomp=np.array([],np.int)#to store the number of M atom in each residue
    mol_in_component=[0 for i in range(cmax)]
    natom=['n' for i in range(cmax)] #number of atoms in each residue
    nformula=['n' for i in range(cmax)]
    Elem=[]#in output this variable refers to atomic symbol(s) of chiral centers in the structure      
    compo=mol.components
    component=[[] for i in range(cmax)]
    occupancy=list()
    #inverted_mol = crystal.symmetric_molecule('-x,-y,-z', (0, 0, 0), force=True)
    #print(inverted_mol)
    meso_check=[False for i in range(cmax)]
    try:
        SG=crystal.spacegroup_number_and_setting[0]# store the SG number
    except (RuntimeError,IndexError) as exp:
        SG="Database Error"
    inverted_mol = crystal.symmetric_molecule('-x,-y,-z', (0, 0, 0), force=True) # creation of the inverted molecules    
    for x, (comp,comp_inv) in enumerate(zip(mol.components,inverted_mol.components)):             
      component[x]=comp     
      
      #Perform a tidy up of the structure adding the unknown bond types and missign H atoms
      #this operation is perfomed for each component - 
      #- it means that if this operation fails for one, the calculation is not stop. 
      #in this version if Fails raised the structure is classified as ERROR
      #In future version a tentative to assign chirality of such structures will be attempt 

      try: 
        self.assignAndAddH(comp, NORMALISE)
        self.assignAndAddH(comp_inv, NORMALISE)
        component[x]=comp

      except (IndexError, RuntimeError) as exc:
        warningc='Could not process structure'
        Fail=True

      natom[x]=len(comp.atoms)
      temp=[]
      r=0
      s=0
      m=0
      ChiAtom=[]
      atom_chirality=[]
      atom_chirality_inv=[]
      
      #loop to count the number and determine the chirality of each atoms in each molecule
      for i,(atom,atom_inv) in enumerate(zip(comp.atoms, comp_inv.atoms)):                            
        neighbours=atom.neighbours
        neighbours_inv=atom_inv.neighbours
        at_sym_bond=[]


        for atom_n,atom_n_inv in zip(neighbours,neighbours_inv):#also save neighbouring atoms of each atom
          at_sym_bond.append(atom_n.atomic_symbol)
        temp.append([str(atom.atomic_symbol), str(at_sym_bond)])

        if atom.is_chiral:
          if atom.occupancy !=1:
            occupancy.append(atom.occupancy)
          if 'H' in atom_n.atomic_symbol:
            H_count+=1#count the number of chiral atom having a H atom as substituent
          if 'C'==atom.atomic_symbol:
            C_count+=1#count the number of chiral carbon atoms
          atom_chirality.append((str(atom.label),atom.chirality))
          Elem.append(str(atom.atomic_symbol)) #keep the chiral atom symbol element in memory
          #print(atom.label,atom.chirality)
          if atom.chirality == 'R':
            r+=1   
          elif atom.chirality == 'S':
            s+=1        
          elif atom.chirality == 'Mixed':
            m+=1
        if atom_inv.is_chiral:
          atom_chirality_inv.append((str(atom_inv.label),atom_inv.chirality))
        
        #set the charge of each atom to 0 - it is used to then generate same SMILES for molecule with different ionic charge      
        if atom.formal_charge != 0:
          atom.formal_charge = 0
          formal_charge=True
            
        for a, b in zip(atom_chirality_inv,atom_chirality):
          #print(a,b)
          if a[1] == b[1]:
            meso_check[x]=True

      # after setting formal_charge to zero, we reassign atom bond type and add hydrogens for each components
      if formal_charge:#if flag true, tentative to retry bond type and H-it do not change chirality bacause it was determined before
        try:
          self.assignAndAddH(comp, NO_NORMALISE)
        except (IndexError, RuntimeError) as exc:
          warningc='Could not process structure'
          Fail=True


      nformula[x]=comp.smiles #the SMILES of the component is save
      mol_in_component[x]=[nformula[x], natom[x], temp, atom_chirality,meso_check[x]]# same list than mol_atom_link completed with a list of chiral atom and absolut configuration in the component
      Rcomp=np.append(Rcomp,r)#save the number of R atoms for each component 
      Scomp=np.append(Scomp,s)#save the number of S atoms for each component 
      Mcomp=np.append(Mcomp,m)#save the number of M atoms for each component 
        

    compo=mol.components
    #variable calculations for the next    
    SumR=sum(Rcomp)#save the total number of R atoms in the structure 
    SumS=sum(Scomp)#save the total number of S atoms in the structure 
    SumM=sum(Mcomp)#save the total number of M atoms in the structure
    pos=np.unique(np.concatenate([np.nonzero(Rcomp)[0],np.nonzero(Scomp)[0]]))#it saves the molecular index chiral molecules
    chiral_mol_N=len(pos)#determine the number of chiral molecules in the structure
    N_form_chi=[nformula[x] for x in pos if nformula[x] is not None]#formula associated to each chiral residue
    Nc=len(sorted(list(set(N_form_chi))))#number of different different chiral chemicals
    
    point_group=['pg'for x in range(Nc)]
    point_group_of_chiral_mol=[]
    #point_group_for_achiral_mol=[]#not yet implemented
    #note_for_achiral_mol=[]
    mol_identical_same_chirality=[[] for x in range(Nc)] #to save pair of molecules having the same chirality
    mol_identical_not_same_chirality=[[] for x in range(Nc)] #to save pair of molecules having the different chirality
    mol_same_chirality_couples_N=[[] for x in range(Nc)]
    mol_not_same_chirality_couples_N=[[] for x in range(Nc)]
    chi_type=[0 for x in range(Nc)]
    chi_type_N=[0 for x in range(Nc)]
    chi_type_x=[[] for x in range(Nc)]
    couple_chirality_N=[0 for x in range(Nc)]
    chi_number=[0 for x in range(Nc)]
    AU_x=['AU_x-error' for x in range(Nc)]#to store the chirality relationship between each chiral family 
    store=[[] for x in range(Nc)]
    couple_chirality=[0 for x in range(Nc)]
    #variables for rmsd comparison


    rmsd_with_inversion=[[] for x in range(Nc)]
    rmsd_same_chirality=[[] for x in range(Nc)]
    rmsd_not_same_chirality=[[] for x in range(Nc)]
    tanimoto_with_inversion=[[] for x in range(Nc)]
    tanimoto_same_chirality=[[] for x in range(Nc)]
    tanimoto_not_same_chirality=[[] for x in range(Nc)]

 
    if Fail:#true if a problem occured during bond type assignment and adding H 
      final="ERROR due to fail"
      AU_x=['AU_x-error']
            
    elif chiral_mol_N==0:#conditions for achiral AU -- No Chiral centre in the structure
      AU="achiral"
      warning="NA"
      final="ACHIRAL"
      Elem=['No Chiral atom']

    elif SumM!=0:#if the structure contains a Miwed type atom, the structure will not be treated
        #the structure could also contain R or S atom, for the moment it not treated    
      warning="contains "+str(sum(Mcomp))+" mixed chiral(s) atom(s)"
      AU_x=['Mixed']
            
    
    rstructMatch = []
    #the following fo loop will scan for each chiral chemical the nature of their relationship
    for m,Fx in enumerate(set(N_form_chi)):
      
      chi_number[m]=N_form_chi.count(Fx)#determine the number molecule having identical atom connectivity 
      #Note that trial to do this operation using atom connectivity was performed
      #But the cannonical atom reordering by the function .normalise_atom_order() creates too much discrepencies
      #in the atom order of neighbours resulting in a too large set of erroneous  molecular mismatch
      
      #determine for a chemical the couples with same chirality and not same chiraity
      tchi=[]
      temp_chi=[]
      #print("pos",pos)
      if chi_number[m]==1:#true if this chemical family contains only one molecule with at least one R or one S atom                        
        rmsd_with_inversion[m]=[-1] #value -1 means that the caluculation is not possible
        tanimoto_with_inversion[m]=[-1]
        
        for i in pos:
          if mol_in_component[i][0] == Fx:
            for c in mol_in_component[i][3]:
              temp_chi.append(c[1])

            tchi.append(temp_chi)  
            chi_type[m]=[''.join(x) for x in tchi]
            chi_type_N[m]=len(list(set(chi_type[m])))
            for ichi,elem in enumerate(set(chi_type[m])):
              chi_type_x[m].append((chi_type[m].count(elem),elem))

            self.c_pointGroup(compo, pos, point_group_of_chiral_mol)

            test = SumR+SumS >= 2
            test2 = (point_group[m] not in chiral_point_group) & (crystal.z_prime<1 if test else True)
            AU_x[m] = "MESO" if test2 else "chiral"
            warning= ("Condition 2" if test else warning) if test2 else ("W1_enantiopure AU" if test else "W2_enantiopure AU")

      else: #true if this chemical family contain more than one molecule in th AU       

        for i, molecule in enumerate(itertools.combinations(pos,2)):#compare each molecule of the same family

          listMol = [molecule[0], molecule[1]]
          mol_a=molecule[0]
          mol_b=molecule[1]
          chi_a=mol_in_component[mol_a][3]
          chi_b=mol_in_component[mol_b][3]
          Ra=Rcomp[mol_a]
          Sa=Scomp[mol_a]
          Rb=Rcomp[mol_b]
          Sb=Scomp[mol_b]
          if mol_in_component[mol_a][0] == mol_in_component[mol_b][0] == Fx:
            #store chiral molecule id belonging to this chemical family
            for i in listMol :
              if i not in store[m] :
                temp_chi=[]
                store[m].append(i)
                for c in (chi_a if i == listMol[0] else chi_b):
                  temp_chi.append(c[1])
                tchi.append(temp_chi)  
                             
            else:#will check if it is enantiomer                   
              for c,d in zip(chi_a,chi_b): #compare the chirality of each atom in globality  
                #true if both molecule have different chirality    
                if c[1]+d[1]=='RS' or c[1]+d[1]=='SR':#couple of counter-enantiomer
                  mol_identical_not_same_chirality[m].append((mol_a,mol_b))
                elif c[1]+d[1]=='RR' or c[1]+d[1]=='SS':#means couple of same enantiomer or stereoisomer
                  mol_identical_same_chirality[m].append((mol_a,mol_b))

            
            mol_identical_same_chirality[m]=list(set(mol_identical_same_chirality[m]))
            mol_identical_not_same_chirality[m]=list(set(mol_identical_not_same_chirality[m]))
            #check if the couple was classified in two subsets (same_chirality and not_same)
            mol_duplicates=[[] for x in range(Nc)]
            
            mol_duplicates[m]=list(set(mol_identical_not_same_chirality[m]).intersection(set(mol_identical_same_chirality[m])))
            if mol_duplicates[m]!=[]:
              mol_identical_same_chirality[m]=[x for x in mol_identical_same_chirality[m] if x not in mol_duplicates[m]]

            mol_same_chirality_couples_N[m]=len(mol_identical_same_chirality[m])
            mol_not_same_chirality_couples_N[m]=len(mol_identical_not_same_chirality[m])
              
          #Point group determination of the set of same molecule base on one of them     

          self.c_pointGroup(compo, store[m], point_group_of_chiral_mol)

          mol_for_overlay=[compo[x] for x in store[m]]


          listOfList = []
          listOfList.append([itertools.combinations(mol_for_overlay,2), rmsd_with_inversion[m], tanimoto_with_inversion[m]])
          if len(mol_identical_same_chirality[m]) > 0 :
            listOfList.append([mol_identical_same_chirality[m], rmsd_same_chirality[m], tanimoto_same_chirality[m]])
          if len(mol_identical_not_same_chirality[m]) > 0 :
            listOfList.append([mol_identical_not_same_chirality[m], rmsd_not_same_chirality[m], tanimoto_not_same_chirality[m]])

          self.c_StructureMatching(listOfList, mol_for_overlay, compo)

        tchi_b=tchi
        tchi=[''.join(x) for x in tchi]
        # chi_type[m]=[''.join(str(x)) for x in tchi]
        # chi_type[m]=[''.join(x) for x in tchi]
        chi_type[m]=tchi
        chi_type_N[m]=len(list(set(chi_type[m])))
          
        for ichi,elem in enumerate(set(chi_type[m])):
          chi_type_x[m].append((chi_type[m].count(elem),elem))
        for el in tchi_b:
            for c_el,chi_el in enumerate(el):
              if chi_el == 'R':
                el[c_el] = 1
              elif chi_el == 'S':
                el[c_el]=-1

        couple_chirality_m=[[] for x in range(len(list(itertools.combinations(list(set(chi_type[m])),2))))]
          
        couple_chirality[m]= couple_chirality_m if chi_type_N[m]>1 else 'C' 


        if chi_type_N[m]>1:
          for n, chir in enumerate(itertools.combinations(list(set(chi_type[m])),2)):#not used in the following
            temp_chi=[]
            for c,d in zip(chir[0],chir[1]):

              temp_chi.append(0 if c + d == 'RS' or c + d=='SR' else 2 if c + d == 'RR' else -2)

            enantiomer_N=np.count_nonzero(temp_chi)
              #Calculation of the relation between the two chiralities

            # couple_chirality_m[n].append('R' if enantiomer_N == 0 else 'C' if Rcomp[mol_a] + Scomp[mol_a] == Rcomp[mol_b] + Scomp[mol_b] and enantiomer_N==len(temp_chi) else 'S')
            couple_chirality_m[n].append('R' if enantiomer_N == 0 else 'C' if Ra + Sa == Rb + Sb and enantiomer_N==len(temp_chi) else 'S')

        couple_chirality_N[m]=len(set(map(tuple,couple_chirality[m])))
              
        
        if point_group[m] not in chiral_point_group :
            AU_x[m]="MESO"
        elif mol_not_same_chirality_couples_N[m] == 0:#all enantiomer have the same chirality
            AU_x[m]="chiral"
            warning="Wa-could be Diast"
        else:#at least 2 enantiomers of different chirality
            if mol_same_chirality_couples_N[m]==0:#no enantiomer of the same chirality 
                if couple_chirality_N[m]>2:
                    AU_x[m]="diast"
                    warning="contains Stereoisomer"
                else:
                    if 'S' in chain(*couple_chirality[m]):
                        AU_x[m]="diast"
                        warning="DIAST-1, contains DIAST"
                    else:
                        if chi_type_N[m]%2==0:
                            AU_x[m]="racemic"#could be either scalemic or stereoisomer
                            warning="RAC-1"
                        else:
                            AU_x[m]="scalemic"#could be either scalemic or stereoisomer
                            warning="SCAL-1"
            else:#contain 2 enantiomers of the same chirality  or stereoisomer or scalemic composition
                if couple_chirality_N[m]>2:
                    
                        AU_x[m]="diast"
                        warning="diast-2 contains Stereoisomer"    
                else:  #chi_type_N<2 there is only 2 kind of stereomers
                    
                    if 'S' in chain(*couple_chirality[m]):
                        AU_x[m]="diast"
                        warning="DIAST-3"
                    else:
                        #print(chi_number[m], chi_type_x[m], chi_type_N[m])
                        if chi_number[m]%2==0:#even number of chiral mol
                            #print(chi_type_x[0][0][0])
                            if chi_type_x[m][0][0]==chi_type_x[m][1][0]:
                                AU_x[m]="racemic"
                                warning="RAC-2"
                            else:
                                AU_x[m]="scalemic"
                                warning="scal-3a-attention could contain stereoisomer"
                        else:
                            AU_x[m]="scalemic"
                            warning="scal-3b-attenton could contain stereoisomer"
          # test = mol_same_chirality_couples_N[m] == 0
          # # if couple_chirality_N[m]>2:
          # if couple_chirality_N[m]>2 and test:
            # AU_x[m]="diast"
            # warning= "contains Stereoisomer" if test else "diast-2 contains Stereoisomer"
          # else:
            # if 'S' in chain(*couple_chirality[m]):
                # AU_x[m]="diast"
                # warning="DIAST-1, contains DIAST" if test else "DIAST-3"
          # test2 = chi_type_N[m] % 2 == 0 if test else chi_number[m] % 2 == 0
          # if test2:
            # AU_x[m]="racemic"#could be either scalemic or stereoisomer
            # warning="RAC-1" if test else "RAC-2"
          # else:
            # AU_x[m]="scalemic"#could be either scalemic or stereoisomer
            # warning="SCAL-1" if test else "scal-3-attention could contain stereoisomer"

    #loop to determine and save the symbol element of atoms having chirality
    for n,chiral_atom_symb in enumerate(Elem):  
      if chiral_atom_symb not in ChiAtom:
        #print ("Elem",chiral_atom_symb)
        ChiAtom.append(chiral_atom_symb)
        #print("ChiAtom",ChiAtom,ChiAtom[0])
        warningb = "contains following chiral atom(s)"+','.join(ChiAtom)
        
    if True in [meso_check[x] for x in pos]:
      AU_x.append("MESO")          
    AU=set((AU_x))#set the AU to reduce the AU chirality value to (at the maximum): [chiral, diast, scalemic, racemic ]
    
    #organized the rmsd values
    listRmsd = [rmsd_with_inversion, rmsd_same_chirality, rmsd_not_same_chirality, tanimoto_with_inversion, tanimoto_same_chirality, tanimoto_not_same_chirality]
    for i in listRmsd :
      listRmsd[listRmsd.index(i)] = [['ND' if len(r) == 0 or not i else r[0],'ND','ND','ND'] if len(r) <= 1 else [round(np.mean(r), NUMBER_AFTER_COMMA),round(np.std(r),NUMBER_AFTER_COMMA), min(r or [0]), max(r or [0])] for r in i] if i else [['ND', 'ND', 'ND', 'ND']]
       
    final_note="NA"

    if 'AU_x-error' in AU:
      final='ERROR'
      final_note="Error occurs in the determination of the AU chirality"
      
    elif 'Mixed' in AU:
      final='ERROR'
      final_note="AU contains mixed chirality type" 

    elif 'ACHIRAL' !=final:
      if "MESO" in AU_x:
        final="MESO"

      elif (not crystal.is_sohncke) & ("scalemic" in AU_x or "racemic" in AU_x or "chiral" in AU_x or "diast" in AU_x):
        final="RACEMIC"

      elif "diast" in AU_x:
        final="DIASTEREOMER"

      else:
        if len(AU)>2:
          final = "Undetermined"
          final_note = "Contains at least 3 different chemicals in enantiopure, racemic and scalemic proportions"
        
        if not ("chiral" in AU and len(AU) == 1):
          # print("condi_1")
          krypto = "KRYPTORACEMIC" 

        if len(AU)==1:
          if "racemic" in AU:
            final = "RACEMIC"
            final_note = "Strict definition of kryptoracemate" if len(AU_x) == 1 else "Contains at least 2 different chemicals in racemic proportions"

          elif "scalemic" in AU:
            final = "SCALEMIC"
            final_note = "Contains 2 different chemicals in scalemic proportions" if len(AU_x) > 1 else "Unbalanced Kryptoracemate"

          elif "chiral" in AU:
            final = "CHIRAL"
            final_note = "Enantiopure" if len(AU_x)==1 else "Contains at least 2 different chemicals in enantiopure proportions"
                                
        elif len(AU)==2:
          if "scalemic" in AU and "racemic" in AU:
            final = "RACEMIC"
            final_note = "contains also an other chemical enantiomer in scalemic proportion"

          elif "scalemic" in AU and "chiral" in AU:
            final = "SCALEMIC"
            final_note = "Unbalanced Kryptoracemate -contains also an other enantiomer in enantiopure proportion" 

          elif "chiral" in AU and "racemic" in AU:
            final = "RACEMIC"
            final_note = "Contains also at least an other enantiomer in enantiopure proportion"

    if chiral_mol_N != cmax:
      final_note = final_note + ' + achiral molecule'

    l = [
      str(e.identifier),
      # str(chi_type_N),
      str(final),
      # str(couple_chirality_N),
      str(final_note),
      str(krypto),
      str(diast),
      str(crystal.spacegroup_symbol),
      str(SG),
      str(crystal.is_sohncke),
      str(e.has_disorder),
      str(occupancy),
      str(round(e.r_factor, NUMBER_AFTER_COMMA) if e.r_factor else e.r_factor),
      str(chiral_mol_N),
      str(Nc),
      str(sum(SumR+SumS)),
      str(sort(ChiAtom)),
      str(C_count),
      str(H_count),
      str(SumR),
      str(SumS),
      str(SumM),
      str(chi_type_x),
      str(chi_type_N),
      str(chi_number),
      str(AU_x),
      str(list(set(point_group_of_chiral_mol))),
      str(cmax),
      str(crystal.z_prime),
      str(crystal.z_value),
      str(warning),
      str(warningb),
      str(warningc),
      self.printArray(crystal.cell_lengths, OTHER),
      self.printArray(crystal.cell_angles, OTHER),
      str(round(crystal.cell_volume, NUMBER_AFTER_COMMA)),
      str(round(crystal.calculated_density, NUMBER_AFTER_COMMA)  if crystal.calculated_density else crystal.calculated_density),
      str(e.formula),
      str(e.chemical_name).encode("utf-8"),
      str(e.temperature),
      str(e.pressure),
      str(e.melting_point),
      str(e.color),
      str(e.polymorph),
      str(e.phase_transition),
      str(e.solvent),
      str(e.remarks),
      self.printArray(e.publication, OTHER),
      str(round((time.time() - start_time), NUMBER_AFTER_COMMA)),
    ]
    if (rstructMatch) :
      for i in rstructMatch:
        l.append(i)
    else :
      for i in range(0,len(sm.StructureMatching().getApproxArray()) * 2):
        l.append("NC")
    for i in l:
      l[l.index(i)] = "\"%s\"" %i
    s = ','.join(l)
    for i in listRmsd :
      s = s + ',' + "\"%s\"" %self.printArray2(i)
    
    # queue.put(s)

    return s

r = FUN_FAILURE
if __name__ == '__main__':
  args = sys.argv
  if((len(args[1]) == "--help") or (len(args[1]) == len("-h"))) & (str(args[1]) ==  "--help" or (str(args[1]) == "-h")):
    Writer().printhelp()
    r = FUN_SUCCESS
  else:
    w = Writer(args[INPUT_FILE])
    r = w.analyse()

exit(r)


